#include <S:\Engine\Graphics\Graphics3D.h>

#include <S:\Engine\Log\Logger.h>

#include <string>

using namespace Microsoft::WRL;

#pragma comment (lib, "d3d12.lib")
#pragma comment (lib, "d3dcompiler.lib")
#pragma comment (lib, "dxgi.lib")

D12Graphics::D12Graphics() {
}

D12Graphics::~D12Graphics() {
}

bool D12Graphics::InitGraphics(HWND &MainWnd) {
	ErrorHR(D3D12CreateDevice(InterfaceList[GPU_Num].Adapter.Get(), InterfaceList[GPU_Num].D3D_Level, _uuidof(ID3D12Device), (void**)&pDevice));

	for (auto &il : InterfaceList) {
		il.Adapter.Get()->Release();
	}

	this->MainWnd = MainWnd;
	Width = static_cast<float> (CurrentOutput[0].Width);
	Height = static_cast<float> (CurrentOutput[0].Height);

	SetAspectRatio(Width / Height);

	Viewport.TopLeftX = 0;
	Viewport.TopLeftY = 0;
	Viewport.Width = Width;
	Viewport.Height = Height;
	Viewport.MinDepth = 0.0f;
	Viewport.MaxDepth = 1.0f;

	ScissorRect.left = 0;
	ScissorRect.top = 0;
	ScissorRect.right = static_cast<LONG>(Width);
	ScissorRect.bottom = static_cast<LONG>(Height);

	D3D12_COMMAND_QUEUE_DESC queueDesc = {};
	queueDesc.Flags = D3D12_COMMAND_QUEUE_FLAG_NONE;
	queueDesc.Type = D3D12_COMMAND_LIST_TYPE_DIRECT;

	ErrorHR(pDevice->CreateCommandQueue(&queueDesc, IID_PPV_ARGS(&pCommandQueue)));

	CreateSwapChains();

	CreateDescriptors();

	ErrorHR(pDevice->CreateCommandList(0, D3D12_COMMAND_LIST_TYPE_DIRECT, pCommandAllocators[FrameIndex].Get(), nullptr, IID_PPV_ARGS(&pCommandList)));

	pCommandList->Close();

	GUI = std::make_unique <D2DGraphics>();

	GUI->InitD2D(pDevice, pCommandQueue, pRenderTargets);

	ErrorHR(pDevice->CreateFence(FenceValues[FrameIndex], D3D12_FENCE_FLAG_NONE, IID_PPV_ARGS(pFence.GetAddressOf())));

	FenceValues[FrameIndex]++;

	HANDLE hFenceEvent = CreateEventEx(nullptr, FALSE, FALSE, EVENT_ALL_ACCESS);
	if (hFenceEvent == nullptr) {
		Logger("CreateEventEx failed");
	}

	return true;
}

void D12Graphics::Destroy() {
	pCommandList.Reset();
	Log_Write("List Released");
	pRtvHeap.Reset();
	Log_Write("RtvHeap Released");
	pCommandQueue.Reset();
	Log_Write("Que Released");
	pSwapChain.Reset();
	Log_Write("Swapchain Released");
	pDevice->Release();
	Log_Write("Device Released");

	for (int i = 0; i < FrameCount; ++i) {
		pRenderTargets[i].Reset();
		Log_Write("RTarged Released");
		pCommandAllocators[i].Reset();
		Log_Write("Allocator Released");
	};

	pFence.Get()->Release();
	Log_Write("Fence Released");

	CloseHandle(hFenceEvent);
}

void D12Graphics::Render() {

	ErrorHR(pCommandAllocators[FrameIndex]->Reset());
	
	PopulateCommandList();

	ID3D12CommandList* ppCommandLists[] = { pCommandList.Get() };
	pCommandQueue->ExecuteCommandLists(_countof(ppCommandLists), ppCommandLists);

	GUI->RenderD2D(FrameIndex);

	ErrorHR(pCommandQueue->Signal(pFence.Get(), FenceValues[FrameIndex]));
	
	ErrorHR(pSwapChain->Present(0, 0));

	MoveToNextFrame();
}

void D12Graphics::PopulateCommandList() {
	HRESULT hr{ S_OK };

	ErrorHR(pCommandList->Reset(pCommandAllocators[FrameIndex].Get(), NULL));
	
	pCommandList->ResourceBarrier(1, &CD3DX12_RESOURCE_BARRIER::Transition(pRenderTargets[FrameIndex].Get(), D3D12_RESOURCE_STATE_PRESENT, D3D12_RESOURCE_STATE_RENDER_TARGET));

	CD3DX12_CPU_DESCRIPTOR_HANDLE rtvHandle(pRtvHeap->GetCPUDescriptorHandleForHeapStart(), FrameIndex, RtvDescriptorSize);

	pCommandList->OMSetRenderTargets(1, &rtvHandle, FALSE, nullptr);

	pCommandList->ClearRenderTargetView(rtvHandle, clearColor, 0, nullptr);

	pCommandList->RSSetViewports(1, &Viewport);
	pCommandList->RSSetScissorRects(1, &ScissorRect);

	pCommandList->ResourceBarrier(1, &CD3DX12_RESOURCE_BARRIER::Transition(pRenderTargets[FrameIndex].Get(), D3D12_RESOURCE_STATE_RENDER_TARGET, D3D12_RESOURCE_STATE_PRESENT));

	ErrorHR(pCommandList->Close());
}

void D12Graphics::MoveToNextFrame() {
	const UINT64 currentFenceValue = FenceValues[FrameIndex];
	ErrorHR(pCommandQueue->Signal(pFence.Get(), currentFenceValue));
	
	FrameIndex = pSwapChain->GetCurrentBackBufferIndex();

	if (pFence->GetCompletedValue() < FenceValues[FrameIndex]) {
		ErrorHR(pFence->SetEventOnCompletion(FenceValues[FrameIndex], hFenceEvent));

		WaitForSingleObjectEx(hFenceEvent, INFINITE, FALSE);
	}

	FenceValues[FrameIndex] = currentFenceValue + 1;
}

bool D12Graphics::CreateDescriptors() {
	D3D12_DESCRIPTOR_HEAP_DESC rtvHeapDesc = {};
	rtvHeapDesc.NumDescriptors = FrameCount;
	rtvHeapDesc.Type = D3D12_DESCRIPTOR_HEAP_TYPE_RTV;
	rtvHeapDesc.Flags = D3D12_DESCRIPTOR_HEAP_FLAG_NONE;

	ErrorHR(pDevice->CreateDescriptorHeap(&rtvHeapDesc, IID_PPV_ARGS(&pRtvHeap)));

	RtvDescriptorSize = pDevice->GetDescriptorHandleIncrementSize(D3D12_DESCRIPTOR_HEAP_TYPE_RTV);

	CD3DX12_CPU_DESCRIPTOR_HANDLE rtvHandle(pRtvHeap->GetCPUDescriptorHandleForHeapStart());

	for (UINT n = 0; n < FrameCount; n++) {
		ErrorHR(pSwapChain->GetBuffer(n, IID_PPV_ARGS(&pRenderTargets[n])));

		pDevice->CreateRenderTargetView(pRenderTargets[n].Get(), nullptr, rtvHandle);

		rtvHandle.Offset(1, RtvDescriptorSize);

		ErrorHR(pDevice->CreateCommandAllocator(D3D12_COMMAND_LIST_TYPE_DIRECT, IID_PPV_ARGS(&pCommandAllocators[n])));
	}

	return true;
}

bool D12Graphics::CreateSwapChains() {
	DXGI_SAMPLE_DESC sampleDesc{};
	sampleDesc.Count = 1;

	DXGI_SWAP_CHAIN_DESC1 swapChainDesc{};
	swapChainDesc.BufferCount = FrameCount;
	swapChainDesc.Width = static_cast<UINT>(Width);
	swapChainDesc.Height = static_cast<UINT>(Height);
	swapChainDesc.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
	swapChainDesc.BufferUsage = DXGI_USAGE_RENDER_TARGET_OUTPUT;
	swapChainDesc.SwapEffect = DXGI_SWAP_EFFECT_FLIP_DISCARD;
	swapChainDesc.SampleDesc = sampleDesc;

	ShowWindow(MainWnd, SW_SHOW);

	ComPtr<IDXGISwapChain1> swapChain{};
	ErrorHR(factory->CreateSwapChainForHwnd(pCommandQueue.Get(), MainWnd, &swapChainDesc, nullptr, nullptr, &swapChain));

	ErrorHR(swapChain.As(&pSwapChain));

	FrameIndex = pSwapChain->GetCurrentBackBufferIndex();

	return true;
}

bool D12Graphics::GetResolutions(Config &cfg) {
	IDXGIOutput* adapterOutput{};

	HRESULT result{};
	int count{};

	while (result = InterfaceList[GPU_Num].Adapter->EnumOutputs(count, &adapterOutput) != DXGI_ERROR_NOT_FOUND) {
		if (FAILED(result)) {
			Log_Write("enum output failed\n");

			return false;
		}

		Screen scr{};

		ErrorHR(adapterOutput->GetDesc(&scr.desc));

		DEVMODE displayMode{};
		displayMode.dmSize = sizeof(DEVMODE);

		if (!EnumDisplaySettings(scr.desc.DeviceName, ENUM_CURRENT_SETTINGS, &displayMode)) {
			Logger("EnumDisplaySettings failed");

			return false;
		}

		float proc = (float)displayMode.dmPelsHeight / (scr.desc.DesktopCoordinates.bottom - scr.desc.DesktopCoordinates.top);

		unsigned int numModes{};
		ErrorHR(adapterOutput->GetDisplayModeList(DXGI_FORMAT_B8G8R8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, NULL));

		DXGI_MODE_DESC* displayModeList = new DXGI_MODE_DESC[numModes];
		if (!displayModeList) {
			Logger("displayModeList allocation failed");

			return false;
		}
		
		auto hr = ErrorHR(adapterOutput->GetDisplayModeList(DXGI_FORMAT_B8G8R8A8_UNORM, DXGI_ENUM_MODES_INTERLACED, &numModes, displayModeList));
		if (FAILED(hr)) { delete[] displayModeList;  return false; }
		
		scr.OutputNum = count;
		scr.DPI = static_cast<int>(proc * 100.0);
		cfg.Feature = InterfaceList[GPU_Num].D3D_Level;

		//if (cfg.Width != displayModeList[cfg.GraphModeNum].Width || cfg.Height != displayModeList[cfg.GraphModeNum].Height) {
		if (cfg.Width < 800 && cfg.Height < 600) {
			scr.ModeNum = numModes - 1;
			scr.Width = displayModeList[scr.ModeNum].Width;
			scr.Height = displayModeList[scr.ModeNum].Height;
			cfg.Width = scr.Width;
			cfg.Height = scr.Height;
			cfg.GraphModeNum = scr.ModeNum;
			Log_Write("Default configuration");
		}
		else {
			scr.Width = cfg.Width;
			scr.Height = cfg.Height;
			Log_Write("Configuration file loaded");
		}

		CurrentOutput.push_back(scr);

		delete[] displayModeList;

		count++;
	}

	return true;
}

void D12Graphics::AdapterList(D12Interface& Current) {
	size_t i{};
	for (auto &e : InterfaceList) {
		std::wcout << i <<L": " << e.GPU_Name.c_str() << std::endl;
		i++;
	}
}

bool D12Graphics::CheckVersionSupport() {
	ID3D12Device* pDevice{};
	ErrorHR(D3D12CreateDevice(InterfaceList[GPU_Num].Adapter.Get(), D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), (void**)&pDevice));
	
	static const D3D_FEATURE_LEVEL fl[] =
	{
		D3D_FEATURE_LEVEL_12_1,
		D3D_FEATURE_LEVEL_12_0,
		D3D_FEATURE_LEVEL_11_1,
		D3D_FEATURE_LEVEL_11_0,
	};

	D3D12_FEATURE_DATA_FEATURE_LEVELS featLevels =
	{
		_countof(fl), fl, D3D_FEATURE_LEVEL_11_0
	};

	ErrorHR(pDevice->CheckFeatureSupport(D3D12_FEATURE_FEATURE_LEVELS, &featLevels, sizeof(featLevels)));

	InterfaceList[GPU_Num].D3D_Level = featLevels.MaxSupportedFeatureLevel;

	switch (featLevels.MaxSupportedFeatureLevel) {
	case D3D_FEATURE_LEVEL_12_1:
		InterfaceList[GPU_Num].strLevel = L"Dx12_1 version\n";
		break;
	case D3D_FEATURE_LEVEL_12_0:
		InterfaceList[GPU_Num].strLevel = L"Dx12_0 version\n";
		break;
	case D3D_FEATURE_LEVEL_11_1:
		InterfaceList[GPU_Num].strLevel = L"Dx11_1 version\n";
		break;
	case D3D_FEATURE_LEVEL_11_0:
		InterfaceList[GPU_Num].strLevel = L"Dx11_0 version\n";
		break;
	}
	
	pDevice->Release();

	return true;
}

bool D12Graphics::CheckSupport() {
	ErrorHR(CreateDXGIFactory1(IID_PPV_ARGS(&factory)));
		
	DXGI_ADAPTER_DESC1 desc;

	int cDevice{ 0 };
	D12Interface Adapt{};

	while (factory->EnumAdapters1(cDevice, &Adapt.Adapter) != DXGI_ERROR_NOT_FOUND) {
		Adapt.Adapter->GetDesc1(&desc);

		if (desc.Flags & DXGI_ADAPTER_FLAG_SOFTWARE)
		{
			factory->EnumWarpAdapter(IID_PPV_ARGS(&Adapt.Adapter));

			ErrorHR(D3D12CreateDevice(Adapt.Adapter.Get(), D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), nullptr));
			
			Adapt.Hardware = false;
			Adapt.GPU_Name = desc.Description;

			InterfaceList.push_back(Adapt);
		}
		else {
			ErrorHR(D3D12CreateDevice(Adapt.Adapter.Get(), D3D_FEATURE_LEVEL_11_0, _uuidof(ID3D12Device), nullptr));

			Adapt.Hardware = true;
			Adapt.GPU_Name = desc.Description;

			InterfaceList.push_back(Adapt);
		}

		ZeroMemory(&Adapt, sizeof(D12Interface));

		cDevice++;
	}

	if (cDevice == 0) return false;

	return true;
}